"""
Load/Unload Problem
===================

Problem originally introduced in `Solving POMDPs by Searching the Space of Finite Policies <https://arxiv.org/pdf/1301.6720.pdf>`_

`Quoting from the original paper on problem description`:

    The load/unload problem with 8 locations: the agent starts in the "Unload" location (U) and receives a reward each time it returns to this place after passing through the "Load" location (L). The problem is partially observable because the agent cannot distinguish the different locations in between Load and Unload, and because it cannot perceive if it is loaded or not (:math:`|S| = 14`, :math:`|O| = 3` and :math:`|A| = 2`).
    
Figure from the paper:

.. figure:: https://i.imgur.com/3OYqqko.png
   :alt: Load/Unload Problem.

   Load/Unload problem


**Not yet implemented**   
"""
