"""
Maze
====

Problem originally introduced in `Solving POMDPs by Searching the Space of Finite Policies <https://arxiv.org/pdf/1301.6720.pdf>`_

`Quoting from the original paper on problem description`:

    A partially observable stochastic maze: the agent must go from the starting state marked with an "S" to the goal marked with an "G". The problem  is partially observable because the agent cannot perceive its true location, but only its orientation and the presence or the absence of a wall on each side of the square defining its current state.  The problem is stochastic because there is a non-zero probability of slipping, so that the agent does not always know if its last attempt to make a move had any consequence on its actual position in the maze.

.. figure:: https://i.imgur.com/i1RDsrL.png
   :alt: Figure from the paper
         
   Maze POMDP


**Not yet implemented**
"""
