"""
Tag
===

For description, see:

* `Point-based value iteration: An anytime algorithm for POMDPs <http://www.fore.robot.cc/papers/Pineau03a.pdf>`_
  The original Tag problem.

* `DESPOT: Online POMDP Planning with Regularization <https://papers.nips.cc/paper/5189-despot-online-pomdp-planning-with-regularization.pdf>`_.
  This includes "Laser Tag", an extension to original Tag.

**Not yet implemented**
"""
